from pyspark.sql.types import StringType, IntegerType
from pipelib.pipeline_components import Step
from pipelib.transformers import *


class PysparkStep(Step):
    """
    Performs multiple transformations over a single PySpark Dataframe, and then merges the results into a single one.

    This Step provides a easy to use interface to configure transformations over a single PySpark Dataframe, runs
    them, and then merges the results using a DataMerger.

    Args:
        transformers: The list of Transformers to be used by the step. Default is an empty list. If you are going to use
            the easy to use interface to configure the transformations, you may want to leave it as its default value.
        data_merger: The DataMerger to be used to merge the results of each transformation into a single result
            DataFrame.
        id_column: The name of the column that contains the id values in the input DataFrame. Some transformations
            need this value to work.
    """
    def __init__(self, data_merger, transformers=None, *, id_column):
        if transformers is None:
            transformers = []
        super().__init__(transformers=transformers, data_merger=data_merger)
        self._transformer_dict = {}
        self.id_column = id_column

    def _get_transformer_types(self):
        return [type(transformer) for transformer in self.transformers]

    def sum_columns(self, list_of_columns, result_name, result_type=StringType()):
        """
        Configures the sum of two or more numeric columns of the input DataFrame.

        Args:
            list_of_columns (List of str): List of columns to be summed
            result_name: Name of the column that will hold the result of the sum.

        Returns:
            This Step, so that method chaining can be used.

        """
        if ColumnsSummator not in self._get_transformer_types():
            columns_adder = ColumnsSummator(id_columns=self.id_column)
            self.transformers.append(columns_adder)
            self._transformer_dict[ColumnsSummator] = columns_adder
        self._transformer_dict[ColumnsSummator].add_sum(list_of_columns=list_of_columns, result_name=result_name)
        return self

    def sum_weighted_columns(self, list_of_columns, list_of_weights, result_name, result_type=StringType()):
        """
        Configures the weighted sum of two or more numeric columns of the input DataFrame.

        Args:
            list_of_columns (List of str): List of columns to be summed
            list_of_weights (List of float): The weight factor of each corresponding column in list_of_columns
            result_name: Name of the column that will hold the result of the sum.

        Returns:
            This Step, so that method chaining can be used.
        """
        if ColumnsWeightedSummator not in self._get_transformer_types():
            columns_weighted_adder = ColumnsWeightedSummator(id_columns=self.id_column)
            self.transformers.append(columns_weighted_adder)
            self._transformer_dict[ColumnsWeightedSummator] = columns_weighted_adder
        self._transformer_dict[ColumnsWeightedSummator].add_weighted_sum(list_of_columns=list_of_columns,
                                                                         list_of_weights=list_of_weights,
                                                                         result_name=result_name)
        return self

    def keep_input_columns(self, columns=None):
        """
        Configures the step to keep some of the columns of the input DataFrame in the Output.

        This is done by a transformation which output is a DataFrame consisting of the specified columns, or if none
        is specified, all columns of the input DataFrame.

        Note:
            If you use this option, you may want to choose a DataMerger that can candle it appropriately.

        Args:
            columns: Columns to be kept from the input dataframe. If none is specified, all columns will be kept.

        Returns:
            This Step, so that method chaining can be used.
        """
        by_pass = ByPass(columns=columns)
        self.transformers.append(by_pass)
        return self
    
    def get_json_field(self, field_path, result_name, result_type=StringType()):
        """
        Configures the step to get a value from a path in a JSON DataFrame.

        Note:
            This method should only be used when the input DataFrame contains JSON fields.

        Args:
            field_path: The path of the value to be get inside the JSON DataFrame.
            result_name: Name of the result column containing the gotten value

        Returns:
            This Step, so that method chaining can be used.
        """
        if JSONFlattener not in self._get_transformer_types():
            json_fattener = JSONFlattener()
            self.transformers.append(json_fattener)
            self._transformer_dict[JSONFlattener] = json_fattener
        self._transformer_dict[ColumnsSummator].add_field(field_path=field_path, result_name=result_name)
        return self

    def binarize(self, column, result_name, result_type=IntegerType()):
        """
        Configures the step to binarize a column of the input DataFrame.

        The binarized column will contain a 0 where the source column were empty, had a 0, an empty string, or any
        value that is evaluated as False by Python. In any other case the value of the binarized column will be 1.

        Args:
            column: The name of the column to be binarized.
            result_name: The name of the result column.

        Returns:
            This Step, so that method chaining can be used.
        """
        if Binarizer not in self._get_transformer_types():
            binarizer = Binarizer(id_columns=self.id_column)
            self.transformers.append(binarizer)
            self._transformer_dict[Binarizer] = binarizer
        self._transformer_dict[ColumnsSummator].add_column_to_binarize(column=column, result_name=result_name)
        return self

    def sum_columns_by_prefix(self, prefix, result_name, result_type=StringType()):
        """
        Configures the step to sum all columns in the input DataFrame which names share a specified prefix.

        Args:
            prefix: The prefix of the name of the columns to be summed.
            result_name: The name of the result column.

        Returns:
            This Step, so that method chaining can be used.
        """
        if ColumnsSummatorByPrefix not in self._get_transformer_types():
            adder_by_prefix = ColumnsSummatorByPrefix(id_columns=self.id_column)
            self.transformers.append(adder_by_prefix)
            self._transformer_dict[ColumnsSummatorByPrefix] = adder_by_prefix
        self._transformer_dict[ColumnsSummatorByPrefix].add_summation(prefix=prefix, result_name=result_name)
        return self

    def extract_value_from_expression(self, source_column, value_to_extract, result_name):
        """
        Configures the step to extract values from a column containing strings of a given structure.

        This transformation generates a column containing values extracted from a source column containing strings
        with a key=pair structure, separated by ';' characters. i.e.: 'key1=value1;key2=value2' and so on.
        More specifically, each cell of the result column will the value of the corresponding cell in the input
        DataFrame, for a single given key.

        Args:
            source_column: The column where the values can be found.
            value_to_extract: The key corresponding to de value to extract.
            result_name: The name of the result column.

        Returns:
            This Step, so that method chaining can be used.
        """
        if ValueExtractor not in self._get_transformer_types():
            extractor = ValueExtractor(id_columns=self.id_column)
            self.transformers.append(extractor)
            self._transformer_dict[ValueExtractor] = extractor
        self._transformer_dict[ValueExtractor].add_extraction(source_column=source_column,
                                                              value_to_extract=value_to_extract,
                                                              result_name=result_name)
        return self

    def count_rows_by_condition(self, sql_condition, result_name, result_type=IntegerType()):
        """
        Configures de step to count the rows that satisfies a given sql condition for each id value in the id columns.

        Args:
            sql_condition: The sql condition to be satisfied by the rows to count.
            result_name: The name of the result column.

        Returns:
            This Step, so that method chaining can be used.
        """
        if RowCounterByCondition not in self._get_transformer_types():
            counter_by_cond = RowCounterByCondition(id_columns=self.id_column)
            self.transformers.append(counter_by_cond)
            self._transformer_dict[RowCounterByCondition] = counter_by_cond
        self._transformer_dict[RowCounterByCondition].add_count(sql_condition=sql_condition, result_name=result_name)
        return self

    def map_values(self, source_column, substitution_dict, default_value, result_name, use_default_value=True, result_type=StringType()):
        """
        Configures the step to do a value mapping of a source column.

        Configures de step to substitute the values of a column using a substitution dict, and then output the results.

        Args:
            source_column: The column which values are to be substituted.
            substitution_dict: A dictionary that specifies the substitutions to perform. Each key-value pair represents
                the original and the substitute value, respectively.
            default_value: The value which will be used to substitute a value that can't be found in the substitution
                dict.
            result_name: The name of the result column.
            use_default_value (bool): Whether the default value should be used. If it's set to false, values that
                can't be found in the substitution dict will remain unchanged.

        Returns:
            This Step, so that method chaining can be used.
        """
        if ValueMapper not in self._get_transformer_types():
            value_mapper = ValueMapper(id_columns=self.id_column)
            self.transformers.append(value_mapper)
            self._transformer_dict[ValueMapper] = value_mapper
        self._transformer_dict[ValueMapper].add_mapping(source_column=source_column, substitution_dict=substitution_dict,
                                                        default_value=default_value, result_name=result_name,
                                                        use_default_value=use_default_value)
        return self

    def get_value(self, source_column, result_name, sql_condition=None, result_type=StringType()):
        """
        Configures the step to generate a column with values of certain rows of the input DataFrame.

        The newly generated column will contain the values of a specific column in the input dataframe, one for each
        row that satisfies a given sql condition.

        Args:
            source_column: The column from which to take the values.
            result_name: The name of the result column.
            sql_condition: The sql condition to be satisfied by the rows containing the desired values.
            result_type:

        Returns:
            This Step, so that method chaining can be used.
        """
        if ValueGetter not in self._get_transformer_types():
            value_finder = ValueGetter(id_columns=self.id_column)
            self.transformers.append(value_finder)
            self._transformer_dict[ValueGetter] = value_finder
        self._transformer_dict[ValueGetter].add_value(value_column=source_column, sql_condition=sql_condition,
                                                      result_name=result_name)
        return self

    def collect_values(self, source_column, result_name, sql_condition=None, result_type=StringType()):
        """
        Configures the step to generate a column containing a collection of certain values for each id.

        The result column will have one row for each id value, containing a list with all the values found on a certain
        column, in rows that satisfy a given sql condition.

        Args:
            source_column: The column from which to take the values.
            result_name: The name of the result column.
            sql_condition: The sql condition to be satisfied by the rows containing the desired values.

        Returns:
            This Step, so that method chaining can be used.
        """
        if ValueCollector not in self._get_transformer_types():
            value_collector = ValueCollector(id_columns=self.id_column)
            self.transformers.append(value_collector)
            self._transformer_dict[ValueCollector] = value_collector
        self._transformer_dict[ValueCollector].add_collecting(value_column=source_column, sql_condition=sql_condition,
                                                              result_name=result_name)
        return self

    def one_hot_encode(self, column_to_encode, result_columns_prefix, sql_condition=None, result_type=StringType()):
        """
        Configures de step to perform one hot encoding to a source column.

        The name of the result columns will be a given prefix followed by the encoded value.

        Args:
            column_to_encode: The column to be one-hot-encoded
            result_columns_prefix: A common prefix to the newly generated columns.
            sql_condition: A optional sql condition to filter the input DataFrame before the encoding.

        Returns:
            This Step, so that method chaining can be used.
        """
        if OneHotEncoder not in self._get_transformer_types():
            one_hot_encoder = OneHotEncoder(id_columns=self.id_column)
            self.transformers.append(one_hot_encoder)
            self._transformer_dict[OneHotEncoder] = one_hot_encoder
        self._transformer_dict[OneHotEncoder].add_enconding(column_to_encode=column_to_encode,
                                                            sql_condition=sql_condition,
                                                            result_prefix=result_columns_prefix)
        return self

    def split_text(self, column_to_split, result_name, delimiter, sql_condition=None, result_type=StringType()):
        """
        Configures de step to split the text found in a column.

        Each cell of the new column will contain a list of strings, in result of the split of the source columns
        string using a given character as delimiter.

        Args:
            column_to_split: The column containing the text to be splitted.
            result_name: The name of the result column.
            delimiter: The delimiter to be used for the text split.
            sql_condition: A optional sql condition to filter the input DataFrame before the split.

        Returns:
            This Step, so that method chaining can be used.
        """
        if TextSplitter not in self._get_transformer_types():
            text_splitter = TextSplitter(id_columns=self.id_column)
            self.transformers.append(text_splitter)
            self._transformer_dict[TextSplitter] = text_splitter
        self._transformer_dict[TextSplitter].add_splitting(source_column=column_to_split, delimiter=delimiter,
                                                           sql_condition=sql_condition, result_name=result_name)
        return self

    def explode_list(self, column_to_explode, result_name, result_type=StringType()):
        """
        Configures the step to explode a column containing lists.

        Generates a new row for each value in each list. The corresponding value for each row will appear in a new
        column.

        Args:
            column_to_explode: The row containing the lists to be exploded.
            result_name: The name of the column that will contain the exploded values.

        Returns:
            This Step, so that method chaining can be used.
        """

        if ListExploder not in self._get_transformer_types():
            list_exploder = ListExploder(id_columns=self.id_column)
            self.transformers.append(list_exploder)
            self._transformer_dict[ListExploder] = list_exploder
        self._transformer_dict[ListExploder].add_exploding(column_to_explode=column_to_explode, result_name=result_name)
        return self

    def do_hashing_trick(self, column_to_hash, number_of_features, result_columns_prefix, result_type=StringType()):
        """
        Does feature hashing trick to a column.

        This is something like the One Hot Enconding, but results in a predefined number of columns.

        .. _More info:
            https://en.wikipedia.org/wiki/Feature_hashing

        Args:
            column_to_hash: The column to hash.
            number_of_features: The number of columns to generate. Should be a power of two.
            result_columns_prefix: A common prefix for the names of the columns to be generated by the feature hashing.

        Returns:
            This Step, so that method chaining can be used.
        """
        if FeatureHasher not in self._get_transformer_types():
            feature_hasher = FeatureHasher(id_columns=self.id_column)
            self.transformers.append(feature_hasher)
            self._transformer_dict[FeatureHasher] = feature_hasher
        self._transformer_dict[FeatureHasher].add_hashing(column_to_hash=column_to_hash, number_of_features=number_of_features,
                                                          result_columns_prefix=result_columns_prefix)
        return self

    def sum_values_by_id(self, column_to_sum, result_name, sql_condition=None, result_type=StringType()):
        """
        Configures the step to sum the values of a column belonging to each id.

        Generates a new column with one row for each different id value, containing the sum of the falues found in a
        numeric column for that value.

        Args:
            column_to_sum: The column containing the numeric values to be summed.
            result_name: The name of the column containing the sum of values for each id.
            sql_condition: An optional sql condition used to filter rows before the sum.

        Returns:
            This Step, so that method chaining can be used.
        """
        if ValueSummatorById not in self._get_transformer_types():
            adder_by_id = ValueSummatorById(id_columns=self.id_column)
            self.transformers.append(adder_by_id)
            self._transformer_dict[ValueSummatorById] = adder_by_id
        self._transformer_dict[ValueSummatorById].add_sum(value_column=column_to_sum, sql_condition=sql_condition,
                                                          result_name=result_name)
        return self
