from pyspark.sql.functions import lit, udf
from pyspark.sql.types import *

def load_csv_cxconnect_extraction_day(data_path, sqlCtx):
    try:
        session_data = sqlCtx.read.load(data_path + "/sessions*",
                                    format='com.databricks.spark.csv',
                                    header=True,
                                    inferSchema = True,
                                    encoding='UTF-8',
                                    delimiter=',')
    except:
        print('Could not read sessions from %s' % data_path)
        session_data = None

    try:
        event_data = sqlCtx.read.load(data_path + "/events*",
                                  format='com.databricks.spark.csv',
                                  header=True,
                                  inferSchema=True,
                                  encoding='UTF-8',
                                  delimiter=',')
    except:
        print('Could not read events from %s' % data_path)
        event_data = None

    try:
        dimension_data = sqlCtx.read.load(data_path + "/report_groups*",
                                   format='com.databricks.spark.csv',
                                   header=True,
                                   inferSchema=True,
                                   encoding='UTF-8',
                                   delimiter=',')
    except:
        print('Could not read report_groups from %s' % data_path)
        dimension_data = None

    return session_data, event_data, dimension_data


class TealeafConnector:
    """
    Merges the three tealeaf datasets (sessions, events and report groups) into a single one.

    Puts all data from sessions, events and report groups datasets into a single one, suited to be used by the
    pipelines library (pipelib).
    """
    def merge(self, sessions_df, events_df, dimensions_df):
        """
        Merges the Tealeaf datasets into a single one.

        Args:
            sessions_df (pyspark.DataFrame): Teleaf sessions data, which must contain the following columns: SESSION_KEY,
                SESSION_DURATION, HIT_COUNT and SESSION_TIMESTAMP
            events_df (pyspark.DataFrame): Tealeaf events data, which must contain the columns: SESSION_KEY, HIT_KEY, EVENT_ID and
                FACT_VALUE
            dimensions_df (pyspark.DataFrame): Tealeaf report groups data, which must contain the columns: SESSION_KEY, EVENT_ID, HIT_KEY,
                GROUP_ID, FACT_DIM_1, FACT_DIM_2, FACT_DIM_3, FACT_DIM_4 and GROUP TYPE

        Returns:
            The merged DataFrame, containing all the columns from the source ones, plus a column ROW_TYPE, showing
                which row came from which data set.
        """
        event_info = events_df.select(["SESSION_KEY", "HIT_KEY", "EVENT_ID", "FACT_VALUE"])
        session_info = sessions_df.select(["SESSION_KEY", "SESSION_DURATION", "HIT_COUNT", "SESSION_TIMESTAMP"])
        dimension_info = dimensions_df.select(
            ["SESSION_KEY", "EVENT_ID", "HIT_KEY", "GROUP_ID", "FACT_DIM_1", "FACT_DIM_2", "FACT_DIM_3", "FACT_DIM_4", "GROUP_TYPE"]
        )

        event_info = (event_info
                      .withColumn('EVENT_TIMESTAMP', lit(None).cast(IntegerType()))
                      .withColumn('FACT_DIM_1', lit(None).cast(StringType()))
                      .withColumn('FACT_DIM_2', lit(None).cast(StringType()))
                      .withColumn('FACT_DIM_3', lit(None).cast(StringType()))
                      .withColumn('FACT_DIM_4', lit(None).cast(StringType()))
                      .withColumn('HIT_COUNT', lit(None).cast(StringType()))
                      .withColumn('SESSION_DURATION', lit(None).cast(StringType()))
                      .withColumn('SESSION_TIMESTAMP', lit(None).cast(StringType()))
                      .withColumn('FACT_ID', lit(None).cast(StringType()))
                      .withColumn('GROUP_ID', lit(None).cast(StringType()))
                      .withColumn('ROW_TYPE', lit("EVENT").cast(StringType()))
                      .withColumn('GROUP_TYPE', lit(None).cast(StringType())))

        def event_timestamp(report_group_type, timestamp):
            if report_group_type != 'TIMESTAMP':
                return None
            return timestamp
        event_timestamp_udf = udf(event_timestamp)

        dimension_info = (dimension_info
                          .withColumn('FACT_VALUE', lit(None).cast(StringType()))
                          .withColumn('EVENT_NAME', lit(None).cast(StringType()))
                          .withColumn('EVENT_TYPE', lit(None).cast(StringType()))
                          .withColumn('HIT_COUNT', lit(None).cast(StringType()))
                          .withColumn('SESSION_DURATION', lit(None).cast(StringType()))
                          .withColumn('SESSION_TIMESTAMP', lit(None).cast(StringType()))
                          .withColumn('ROW_TYPE', lit("REPORT_GROUP").cast(StringType()))
                          .withColumn('EVENT_TIMESTAMP', event_timestamp_udf('GROUP_TYPE', 'FACT_DIM_1')))

        session_info = (session_info
                        .withColumn('FACT_VALUE', lit(None).cast(StringType()))
                        .withColumn('EVENT_TIMESTAMP', lit(None).cast(IntegerType()))
                        .withColumn('FACT_DIM_1', lit(None).cast(StringType()))
                        .withColumn('FACT_DIM_2', lit(None).cast(StringType()))
                        .withColumn('FACT_DIM_3', lit(None).cast(StringType()))
                        .withColumn('FACT_DIM_4', lit(None).cast(StringType()))
                        .withColumn('HIT_KEY', lit(None).cast(StringType()))
                        .withColumn('EVENT_ID', lit(None).cast(StringType()))
                        .withColumn('EVENT_NAME', lit(None).cast(StringType()))
                        .withColumn('FACT_ID', lit(None).cast(StringType()))
                        .withColumn('GROUP_ID', lit(None).cast(StringType()))
                        .withColumn('ROW_TYPE', lit("SESSION").cast(StringType()))
                        .withColumn('GROUP_TYPE', lit(None).cast(StringType())))

        unified_dataset_schema = ['SESSION_KEY', 'SESSION_TIMESTAMP', 'SESSION_DURATION', 'HIT_COUNT', 'ROW_TYPE',
                                  'HIT_KEY', 'EVENT_ID', 'FACT_VALUE', 'GROUP_ID', 'GROUP_TYPE', 'FACT_DIM_1', 'FACT_DIM_2',
                                  'FACT_DIM_3', 'FACT_DIM_4', 'EVENT_TIMESTAMP']

        session_info = session_info.select(unified_dataset_schema)
        dimension_info = dimension_info.select(unified_dataset_schema)
        event_info = event_info.select(unified_dataset_schema)

        merged_data = session_info.union(dimension_info).union(event_info)

        return merged_data
